"""
Plot stuff
"""
from cycler import cycler
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

def plot_speaker_mapping(df, out_path="test/result/speaker-mapping.png"):
    """
    Plot mapped vs unmapped speaker introductions

    Args

        d: dictionary prouced by test.known-speakers
        out_path: where the plot gets written

    Return

        nothing, but writes a plot
    """
    estates = df["estate"].unique()
    print(estates)
    colors = list('kkbbggrrcc')
    default_cycler = (cycler(color=colors) +
                      cycler(linestyle=(['-', '--']*5)) +
                      cycler(linewidth=([2, 1.25]*5))
                      )
    plt.rc('axes', prop_cycle=default_cycler)
    f, ax = plt.subplots(figsize=(16,7))
    legend_text = []
    ddf = df.groupby("year")[["total", "matched"]].sum().reset_index()
    x = ddf['year'].tolist()
    for count in ["total", "matched"]:
        legend_text.append(f"ALL:{count}")
        Y = ddf[count].tolist()
        X, Y = zip(*sorted(zip(x,Y),key=lambda x:x[0]))
        plt.plot(X,Y)
    for estate in estates:
        print(estate)
        dfv = df.loc[df["estate"] == estate]
        x = dfv['year'].tolist()
        for count in ["total", "matched"]:
            legend_text.append(f"{estate}:{count}")
            Y = dfv[count].tolist()
            X, Y = zip(*sorted(zip(x,Y),key=lambda x:x[0]))
            plt.plot(X,Y)
    plt.title('Coverage of matched speakers vs total speakers')
    plt.legend(legend_text, loc ="upper right")
    ax.set_xlabel('Year')
    ax.tick_params(axis='x', labelrotation=90)
    plt.savefig(out_path, dpi=300)



def plot_speaker_mapping_proportion(df, out_path="test/result/speaker-mapping-proportion.png"):
    """
    Plot mapped vs unmapped speaker introductions

    Args

        d: dictionary prouced by test.known-speakers
        out_path: where the plot gets written

    Return

        nothing, but writes a plot
    """
    estates = df["estate"].unique()
    print(estates)
    colors = list('kbgrc')
    default_cycler = (cycler(color=colors) +
                      cycler(linestyle=(['-', '--', '--', '--', '--'])) +
                      cycler(linewidth=([2, 1.25, 1.25, 1.25, 1.25]))
                      )
    plt.rc('axes', prop_cycle=default_cycler)
    f, ax = plt.subplots(figsize=(16,7))
    ddf = df.groupby("year")[["total", "matched"]].sum().reset_index()
    lab = ["total"]
    ddf["proportion"] = ddf.apply(lambda x: x["matched"]/x["total"], axis=1)
    x = ddf['year'].tolist()
    y = ddf['proportion'].tolist()
    X, Y = zip(*sorted(zip(x,y),key=lambda x:x[0]))
    plt.plot(X,Y)
    for estate in estates:
        lab.append(estate)
        dfv = df.loc[df["estate"] == estate]
        dfv["proportion"] = dfv.apply(lambda x: x["matched"]/x["total"], axis=1)
        x = dfv['year'].tolist()
        y = dfv['proportion'].tolist()
        X, Y = zip(*sorted(zip(x,y),key=lambda x:x[0]))
        plt.plot(X,Y)
    plt.title('Coverage of matched speakers vs total speakers')
    plt.legend(lab, loc ="upper right")
    ax.set_xlabel('Year')
    ax.tick_params(axis='x', labelrotation=90)
    plt.savefig(out_path, dpi=300)



"""
Utilities for the Valtiopaivat Corpus
"""
from lxml import etree
from pyriksdagen.utils import (
    elem_iter,
    get_formatted_uuid,
    parse_tei,

)
import json




def write_tei(elem, dest_path, padding=8) -> None:
    """
    Write a corpus document to disk.

    Args:
        elem (etree._Element): tei root element
        dest_path (str): protocol path
    """
    def _sort_attrs(elem):
        custom_order = ["xml:id", "type", "subtype"]
        attrs = sorted(elem.attrib.items())
        if len(attrs) == 0 or type(attrs) == list:
            return elem
        d = {}
        for _ in custom_order:
            if _ in attrs:
                d[_] = attrs[_]
                del attrs[_]
        for k,v in attrs.items():
            d[k] = v
        elem.attrib.clear()
        for k, v in d.items():
            elem.attrib[k] = v
        return elem

    def _iter(root, ns="{http://www.tei-c.org/ns/1.0}"):
        for body in root.findall(".//" + ns + "body"):
            for div in body.findall(ns + "div"):
                for ix, elem in enumerate(div):
                    elem = _sort_attrs(elem)
                    if elem.tag == ns + "u":
                        yield "u", elem
                    elif elem.tag == ns + "note":
                        yield "note", elem
                    elif elem.tag == ns + "pb":
                        yield "pb", elem
                    elif elem.tag == "pb":
                        elem.tag = ns + "pb"
                        yield "pb", elem
                    elif elem.tag == ns + "seg":
                        yield "seg", elem
                    elif elem.tag == "u":
                        elem.tag = ns + "u"
                        yield "u", elem
                    elif elem.tag == "p":
                        elem.tag = ns+"p"
                        yield "p", elem
                    elif elem.tag == ns + "p":
                        yield "p", elem
                    else:
                        warnings.warn(f"Unrecognized element {elem.tag}")
                        yield None

    def _format_paragraph(paragraph, spaces):
        s = "\n" + " " * spaces
        words = paragraph.replace("\n", "").strip().split()
        row = ""
        for word in words:
            if len(row) > 60:
                s += row.strip() + "\n" + " " * spaces
                row = word
            else:
                row += " " + word

        if len(row.strip()) > 0:
            s += row.strip() + "\n" + " " * (spaces - 2)
        if s.strip() == "":
            return None
        return s

    def _format_texts(root, padding=12):
        for tag, elem in _iter(root):
            if tag in ["note", "p"]:
                if type(elem.text) == str:
                    elem.text = _format_paragraph(elem.text, padding+2)
                else:
                    elem.text = None
                if elem.text == None:
                    elem.getparent().remove(elem)
            elif tag == "u":
                if len("".join(elem.itertext())) == 0:
                    elem.getparent().remove(elem)
                elif len(list(elem)) > 0:
                    # Format segs' text content
                    # Remove segs with no text content
                    for seg in elem:
                        if type(seg.text) == str:
                            seg.text = _format_paragraph(seg.text, padding+4)
                        else:
                            seg.text = None
                        if seg.text is None:
                            seg.getparent().remove(seg)
                    elem.text = None
                else:
                    elem.getparent().remove(elem)
        return root

    elem = _format_texts(elem, padding=padding)
    b = etree.tostring(
        elem,
        pretty_print=True,
        encoding="utf-8",
        xml_declaration=True
    )
    with open(dest_path, "wb") as f:
        f.write(b)


def infer_metadata(filename, verbose=False):
    """
    Heuristically infer metadata from a protocol id or filename.

    Args

        filename (str): the protocols filename. Agnostic wrt. dashes and underscores. Can be relative or absolute.
        verbose (bool): print stuff

    Returns a dict with keys "filename", "doc_type", "chamber", "year", and "number"
    """
    metadata = dict()
    doc = filename.split("/")[-1].split(".")[0]
    metadata["filename"] = doc
    doctype, year, chamber, num  = doc.split("_")
    metadata["document_type"] = doctype
    if chamber == "":
        metadata["chamber"] = None
    else:
        metadata["chamber"] = chamber.capitalize()
    metadata['yearstr'] = year
    metadata["year"] = year[:4]
    if len(year) > 4:
        metadata["secondary_year"] = year[-4:]
    else:
        metadata["secondary_year"] = None
    metadata["number"] = num
    if verbose: print("INFO:\n", json.dumps(metadata, indent=2))
    return metadata


XML_NS = "{http://www.w3.org/XML/1998/namespace}"
TEI_NS = "{http://www.tei-c.org/ns/1.0}"



"""
A configuration helper for the Valtiopaivat Corpus.
"""
import json
import os
import warnings




class ValtiopaivatCorpusConfig:
    """
    Config options for the Valtiopaivat Corpus.
    """
    def __init__(self, **kwargs):
        self.ConfigName = None
        self.ConfigPath = None
        self.ValtiopaivatRecordsTEILocation = None
        self.ValtiopaivatRecordsALTOLocation = None
        self.ValtiopaivatRecordsPDFLocation = None
        self.ValtiopaivatRecordsLOMap = None
        self.ValtiopaivatHandlingarTEILocation = None
        self.ValtiopaivatHandlingarALTOLocation = None
        self.ValtiopaivatHandlingarPDFLocation = None
        self.ValtiopaivatHandlingarLOMap = None
        self.ValtiopaivatRegistersTEILocation = None
        self.ValtiopaivatRegistersALTOLocation = None
        self.ValtiopaivatRegistersPDFLocation = None
        self.ValtiopaivatRegistersLOMap = None

        for k,v in kwargs.items():
            if hasattr(self, k):
                setattr(self, k, v)
            else:
                warnings.warn(f"The property -- {k} -- from the config file is not a valid property. Ignoring")

    def write(self):
        """
        Write the config to a file
        """
        with open(self.ConfigPath, 'w+') as outf:
            json.dump(self.__dict__, outf, indent=2)

    def update(self, **kwargs):
        """
        Update attrib values and write

        Args

            kwargs: key = val pairs to update
        """
        for k,v in kwargs.items():
            if hasattr(self, k):
                setattr(self, k, v)
            else:
                warnings.warm(f"The property -- {k} -- from the config file is not a valid property. Ignoring")
        self.write()




def track_existing_config(name = None, location = None, overwrite_existing=False):
    """
    Assign a name to an existing config file.

    Args

        name: a name for the config
        location: path and filename for the saved config

    Return

        bool (True == OK)
    """
    if name == None or location == None:
        raise Exception("You have to explicitly set the 'name' and 'location' of a new config.")

    try:
        with open(f"{os.path.abspath(os.path.dirname(__file__))}/cfg_list.json", 'r') as cfg_list_file:
            cfg_list = json.load(cfg_list_file)
    except:
        cfg_list = {}

    if name in cfg_list:
        if overwrite_existing == False:
            raise Exception(f"The name -- {name} -- already exists as a named config. (pass overwrite_existing=True to override this error)")

    cfg_list[name] = os.path.abspath(location)

    with open(f"{os.path.abspath(os.path.dirname(__file__))}/cfg_list.json", 'w+') as cfg_list_file:
        json.dump(cfg_list, cfg_list_file, indent=2)

    return True


def create_new_config(name = None, location = None, **kwargs):
    """
    Create and save a new config object ans save it to a file.

    Args

        name: a name for the config
        location: path and filename for the saved config
        **kwargs: key = val pairs to initialize with

    Rrturn

        config object
    """
    if track_existing_config(name=name, location=location):
        if kwargs is None:
            kwarks = {}
        kwargs["ConfigName"] = name
        kwargs["ConfigPath"] = os.path.abspath(location)
        cfg = ValtiopaivatCorpusConfig(**kwargs)
        cfg.write()

    return cfg


def load_config(name = None, location = None):
    """
    Load an existing config from file. If the config wis created with create_new_config, it can be loaded by name, otherwise, pass a file path.

    Args

        name: name of config
        location: path and file name of saved config.

    Return

        config
    """
    if name == None and location == name:
        raise Error("You need to pass either a name or location for the config file you want to load")

    if location is None:
        try:
            with open(f"{os.path.abspath(os.path.dirname(__file__))}/cfg_list.json", 'r') as inf:
                cfg_list = json.load(inf)
                location = cfg_list[name]
        except Exception as e:
            print("oops...something went wrong")
            print(e)

    with open(location, 'r') as cfg_file:
        cfg = ValtiopaivatCorpusConfig(**json.load(cfg_file))

    cfg.write()

    return cfg



import os
import regex as re

def compile_pattern(query):
    """
    load query from a file

    Args

        query (str): name of .rq file stored with this module under `regex/`
    """
    if not os.path.exists(f"{os.path.dirname(os.path.abspath(__file__))}/rq/{query}.rq"):
        raise FileNotFoundError(f"Couldn't find file : {os.path.dirname(os.path.abspath(__file__))}/rq/{query}.rq")

    with open(f"{os.path.dirname(os.path.abspath(__file__))}/rq/{query}.rq", 'r') as inf:
        lines = [rf"{_.strip()}" for _ in inf.readlines()]
    q = ''.join(lines)
    #q = q.replace("\\s", r"\s")
    q = q.replace("\\\\","\\")
    print(q)
    pat =  re.compile(rf"{q}")
    print(pat)
    return pat



"""
Functions for curating the Valtiopaivat Corpus from scanned, OCRed image files
"""
from lxml import etree
from pyparlaclarin.create import pc_header
from pyriksdagen.download import _alto_extract_paragraphs
from valtiopy.utils import (
    get_formatted_uuid,
    parse_tei,
    TEI_NS,
    write_tei,
    XML_NS,
)
import alto
import copy
import os




def convert_alto(files):
    """
    Convert a list of alto files to a dict
    {page_index: [paragraph, paragraph, paragraph...]

    NB. There was a graphic type element in the alto xml that was causing the alto package to throw errors.
    I went into the soutce code of the package and changed the line with raise Error...
    to warnings.warn...

    Args

        files: (list) collection of alto file paths

    Return

        paragraphs (dict)
    """
    paragraphs = {}
    in_sync = True
    for file_ in files:

        nr = file_.split('-')[-1].replace('.xml', '')
        try:
            altofile = alto.parse_file(file_)
        except:
            print("OOOPS", file_)
            raise Error("sth isn't right")
        pp = _alto_extract_paragraphs(altofile)
        paragraphs[nr] = pp
    return paragraphs


def dict_to_tei(data, verbose=False):
    """
    Convert a metadata dict into a TEI XML tree

    Args

        data (dict): dictionary containing protocol level metadata
        verbose (bool): print stuff

    Return

        tei (lxml.etree.Element): the protocol as a TEI-formatted lxml tree root
    """
    if verbose: ptint(f"INFO: Preparing tei")
    metadata = copy.deepcopy(data)
    dt = {
        "prot": "records",
        "ptk": "records",
        "hand": "handlingar",
        "ask": "handlingar",
        "bil": "handlingar",
        "reg": "register",
        "sis": "register"
    }
    nsmap = {None: TEI_NS}
    nsmap = {key: value.replace("{", "").replace("}", "") for key,value in nsmap.items()}
    tei = etree.Element("TEI", nsmap=nsmap)
    #protocol_id = metadata["protocol_id"]
    metadata["document_title"] = (
        f"Finnish parliamentary document: {metadata['document_type']} {metadata['yearstr']}, {metadata['chamber']} number {metadata['number']}"
    )
    documentHeader = pc_header(metadata)
    tei.append(documentHeader)

    text = etree.SubElement(tei, "text")
    front = etree.SubElement(text, "front")
    preface = etree.SubElement(front, "div", type="preface")
    etree.SubElement(preface, "head").text = metadata["filename"]
    if "date" not in metadata:
        year = metadata.get("year", 2020)
        metadata["date"] = str(year) + "-01-01"

    etree.SubElement(preface, "docDate", when=metadata["date"]).text = metadata.get(
        "date", "2020-01-01"
    )

    body = etree.SubElement(text, "body")
    body_div = etree.SubElement(body, "div")

    protocol_id = metadata["filename"]
    element_seed = f"{protocol_id}\nNA\n"
    print(element_seed)
    for nr, pp in data["paragraphs"].items():
        pb = etree.SubElement(body_div, "pb")
        pb.attrib["facs"] = f"https://swerik-project.github.io/valtiopaivat-{dt[metadata['document_type']]}-pdf/{metadata['yearstr']}/{metadata['filename']}-{nr}.pdf"
        for paragraph in pp:
            if metadata["document_type"] in ["ptk", "prot"]:
                elem = etree.SubElement(body_div, "note")
            else:
                elem = etree.SubElement(body_div, "p")
            elem.text = paragraph
            element_seed += paragraph
            elem.attrib[f"{XML_NS}id"] = get_formatted_uuid(element_seed)

    return tei



def dict_to_parlaclarin(data, tei_loc, verbose=False):
    """
    Create per-protocol parlaclarin files of all files provided in file_db.
    Does not return anything, instead writes the data on disk.

    Args:

        data (dict): metadata and data
        tei_loc (str): path to tei
        verbose (bool): print stuff
    """
    parlaclarin_path = f"{tei_loc}/{data['filename']}.xml"
    if verbose: print(f"INFO: preparing to write tei to {parlaclarin_path}")
    tei = dict_to_tei(data)
    if verbose: print("INFO: TEI OK... write and read-write test")
    os.makedirs(tei_loc, exist_ok=True)
    write_tei(tei, parlaclarin_path)
    root, ns = parse_tei(parlaclarin_path)
    write_tei(root, parlaclarin_path)
    if verbose: print("INFO:    OK")



from glob import glob
import pandas as pd
import warnings



def fetch_metadata_tables(metadata_path="valtiopaivat-persons/data"):
    """
    return a list of metadata tables.

    Args

        metadata_path (str): place to look for metadata tables

    returns

        list of pd.DataFrame objects
    """
    tables = []
    csv_files = glob(f"{metadata_path}/*.csv")
    for csv in csv_files:
        tables.append(pd.read_csv(csv))
    return tables


def join_metadata_tables(tables, key="swerik_person_id"):
    """
    Join metadata tables on a key. Return all possible combinations of data on the matching key.

    Args

        tables (list): list of pd dataFrame objects
        key (str): common column to merge on

    returns

        pandas df
    """
    assert len(tables) > 0
    if not isinstance(tables, list):
        raise TypeError("Expected a list")
    try:
        assert all(isinstance(_,pd.DataFrame) for _ in tables)
    except Exceptoin as e:
        print("expected a pandas dataframe", e)

    if len(tables) == 1:
        warnings.warn("I can't merge a table with itself. You get back what you put in.")
        return tables[0]
    else:
        inner_df = pd.merge(tables[0], tables[1], how='inner', on=key)
        outer_df = pd.merge(tables[0], tables[1], how='outer', on=key)
        if len(tables) > 2:
            for table in tables[2:]:
                inner_df = pd.merge(inner_df, table, how='inner', on=key)
                outer_df = pd.merge(outer_df, table, how='outer', on=key)
        return pd.concat([inner_df, outer_df]).drop_duplicates()


def fetch_metadata(metadata_path="valtiopaivat-persons/data"):
    """
    Get all metadata tables at metadata path and merge them to one df
    """
    return join_metadata_tables(fetch_metadata_tables(metadata_path=metadata_path)).sort_values(by=["swerik_person_id"])



import hashlib
import numpy as np
import os
import pandas as pd




def goldstandard(df, n=3, by=["year", "estate"], seed=None, scope="dir", sampled_format="pdf"):
    """
    Draw a goldstandard sample of N documents per stratum

    Args

        - df (pandas DataFrame): DF from which to draw a sample. It should have a column `path` and a column for each of the strata
        - n (int): number of items to draw
        - by (list): list of strata to stratify with
        - seed (str): a random state
        - scope (str): "dir" or "file". If path is a `dir`, this will sample from all files in the directory. if scope is `file` this will sample the paths in the path column
        - sampled_format (str): this function uses the alto xml files to draw a sample (easier to store all locally), but often we want PDF files. If this option is set to "pdf" the path will be corrected in the resulting sample to the PDF

    Return

        - list
    """

    #for stratum in strata:
    #    print(stratum)
    #    df_stratum = df[df[by] == stratum]
       #data.append(df_strata.sample(n, random_state=random_state))
        #return pd.concat(data).reset_index(drop=True)

    def _sample_dir(x, sample_, n, random_state=None):
        files_ = []
        x = list(x["path"])
        for _ in x:
            files_.extend([f"{_}/{file_}" for file_ in os.listdir(_)])
        s = pd.Series(files_).sample(n=n, random_state=random_state).to_list()
        sample_.extend(s)
        return sample_

    if seed is not None:
        seed = int(hashlib.sha256(seed.encode("utf-8")).hexdigest(), 16) % (2**32)
    if scope == 'file':
        sample_ = df.groupby(by, group_keys=False).apply(lambda x: x.sample(n, random_state=seed), include_groups=False)
        sample_ = list(sample_["path"])
    elif scope == 'dir':
        sampled_ = []
        sample_ = df.groupby(by, group_keys=False).apply(lambda x: _sample_dir(x, sampled_, n, random_state=seed), include_groups=False)
        sample_=sampled_

    if sampled_format == "pdf":
        sample_ = [_.replace("-alto", "-pdf").replace(".xml", ".pdf") for _ in sample_]
    return sample_



if __name__ == '__main__':
    rows1 = [
        ["file1", "1992", "a"],
        ["file2", "1992", "b"],
        ["file3", "1992", "c"],
        ["file4", "1993", "a"],
        ["file11", "1993", "b"],
        ["file5", "1994", "a"],
        ["file12", "1994", "b"],
        ["file6", "1995", "a"],
        ["file7", "1995", "b"],
        ["file8", "1995", "c"],
        ["file9", "1996", "b"],
        ["file13", "1996", "a"],
        ["file10", "1997", "c"],
        ["file14", "1997", "b"],
        ["file15", "1997", "a"],
        ["file11", "1992", "a"],
        ["file12", "1992", "b"],
        ["file113", "1992", "c"],
        ["file114", "1993", "a"],
        ["file1111", "1993", "b"],
        ["file115", "1994", "a"],
        ["file112", "1994", "b"],
        ["file16", "1995", "a"],
        ["file17", "1995", "b"],
        ["file18", "1995", "c"],
        ["file19", "1996", "b"],
        ["file1113", "1996", "a"],
        ["file111110", "1997", "c"],
        ["file111114", "1997", "b"],
        ["file151111", "1997", "a"],
        ["file17", "1992", "a"],
        ["file27", "1992", "b"],
        ["file37", "1992", "c"],
        ["file47", "1993", "a"],
        ["file171", "1993", "b"],
        ["file57", "1994", "a"],
        ["file127", "1994", "b"],
        ["file67", "1995", "a"],
        ["file77", "1995", "b"],
        ["file87", "1995", "c"],
        ["file97", "1996", "b"],
        ["file173", "1996", "a"],
        ["file107", "1997", "c"],
        ["file147", "1997", "b"],
        ["file157", "1997", "a"],
    ]

    rows2 = [
        ['valtiopaivat-records-alto/data/1882/ptk_1882_talonpojat_I', '1882', 'talonpojat'],
        ['valtiopaivat-records-alto/data/1882/ptk_1882_talonpojat_II', '1882', 'talonpojat'],
        ['valtiopaivat-records-alto/data/1900/prot_1900_adeln_lll', "1900", "adeln"],
        ['valtiopaivat-records-alto/data/1900/prot_1900_adeln_II', "1900", "adeln"],
        ['valtiopaivat-records-alto/data/1900/prot_1900_adeln_I', "1900", "adeln"],
        ['valtiopaivat-records-alto/data/1877-1878/prot_1877-1878_borgare_IV', '1877-1878', 'borgare'],
        ['valtiopaivat-records-alto/data/1877-1878/prot_1877-1878_borgare_III', '1877-1878', 'borgare'],
        ['valtiopaivat-records-alto/data/1877-1878/prot_1877-1878_borgare_II', '1877-1878', 'borgare'],
        ['valtiopaivat-records-alto/data/1877-1878/prot_1877-1878_borgare_I', '1877-1878', 'borgare'],
        ['valtiopaivat-records-alto/data/1877/prot_1877_adeln_II', "1877", "adeln"],
        ['valtiopaivat-records-alto/data/1885/prot_1885_borgare_I', "1885", "borgare"],
    ]
    cols = ["path", "year", "estate"]
    df1 = pd.DataFrame(rows1, columns=cols)
    df2 = pd.DataFrame(rows2, columns=cols)

    print("Running goldstandard test 1: sample files")

    [print(_) for _ in goldstandard(df1, n=1, scope="file")]

    print("Running goldstandard test 2: sample files from dir")

    [print(_) for _ in goldstandard(df2, n=3, scope="dir")]



"""
A DRY argparse helper for common arguments
"""
from glob import glob
from valtiopy.config import (
    create_new_config,
    load_config,
)
import argparse
import warnings




class NotImplemented(Warning):
    """
    Warn on unimplemented functions
    """
    def __init__(self, func):
        self.message = f"\n    --~~--~~>    The function {func} is not implemented yet."

    def __str__(self):
        return self.message


def custom_serializer(obj):
    """
    This takes an argparse args class and either serializes it to txt,
    or replaces a non-serializable object with its name.

    used for debugging the argparse args
    """
    try:
        return json.JSONEncoder().default(obj)
    except TypeError:
        return f"Non-serializable: {type(obj).__name__}"


def populate_common_args(parser):
    """
    Add common arguments to a parser that will be used in most of the scripts.

    Args

        parser: argparse parser instance

    Returns

        parser
    """
    parser.add_argument("-c", "--config-name",
                        default = 'default',
                        help = "Load a config file created with the valtiopy.config module or tracked with it by name")
    parser.add_argument("--config-path",
                        default = None,
                        help = "Point to a config file to load.")
    parser.add_argument("--config-create",
                        action = 'store_true',
                        help = "Create a new config. The config name and config path arguments must be set with this option.")
    parser.add_argument("-s", "--start",
                        type = str,
                        help = "year to start processing")
    parser.add_argument("-e", "--end",
                        type = str,
                        help = "year to end processing")
    parser.add_argument("-m", "--meeting",
                        type = str,
                        help = "A particular meeting year or riksmöte (either a year or range when a given parliament was active).")
    parser.add_argument("-l", "--languages",
                        choices = ["f", "s", "fs"],
                        nargs= "+",
                        default = ['s'],
                        help = "Parse documents with the selected primary languages")
    parser.add_argument("-k", "--chambers",
                        choices = ["adeln", "borgare", "praster", "talonpajat", "papisto", "porvaristo"],
                        nargs = "+",
                        help = "Parse documents from chambers.")
    parser.add_argument("-d", "--doctypes",
                        choices = ["ask", "hand", "prot", "ptk", "bil", "reg", "sis"],
                        nargs = "+",
                        help = "Document types to process")
    parser.add_argument("-f", "--docformats",
                        choices = ["alto", "tei", "pdf"],
                        default = ['tei'],
                        nargs = "+",
                        help = "Which kind of documents do you want to call up?")
    parser.add_argument("-v", "--verbose",
                        action = 'store_true',
                        help = "Print extra information about what's going on")
    return parser


def fetch_parser(description=None):
    """
    Create a parser instance and populate it with common arguments.

    Args

        description (str): description of script calling parser (__doc__)

    Returns

        parser
    """
    parser = argparse.ArgumentParser(description=description)
    parser = populate_common_args(parser)
    return parser


def impute_arg_values(args):
    """
    parse common arguments and pre-handle args for the script user

    Args

        args: Argparse argument class instances

    Returns

        args
    """
    def _handle_config(args):
        """
        fetch or create configif args.verbose: print(f"INFO:
        """
        if args.config_create:
            if args.verbose: print(f"INFO: Creating a new config file called \"{args.config_name}\" at {args.config_path}")
            args.config = create_new_config(name=args.config_name, location=args.config_path)
        else:
            if args.config_path is not None:
                if args.verbose: print(f"INFO: Loading config from path {args.config_path}")
                args.config = load_config(location=args.config_path)
            else:
                if args.verbose: print(f"INFO: Loading config from name {args.config_name}")
                args.config = load_config(name=args.config_name)
        return args

    def _fetch_documents(args):
        """
        fetch documents according to start/end or meeting time frame

        Args

            args: Argparse argument class instances

        Returns

            args
        """
        def _get_files(paths, args, ext=".xml"):
            files = []
            for path_ in paths:
                if args.verbose: print(f"INFO: looking for files at {path_}")
                if args.meeting is not None:
                    if args.verbose: print(f"INFO: riksmöte meeting specified {args.meeting}")
                    fs = sorted(glob(f"{path_}/data/{args.meeting}/**/*{ext}", recursive=True))
                    files.extend(fs)
                    if args.verbose: print(f"INFO:   found {len(fs)} files")
                else:
                    fs = sorted(glob(f"{path_}/data/**/*{ext}", recursive=True))
                    if args.verbose: print(f"INFO: found {len(fs)} files")
                    if args.start is not None:
                        if args.verbose: print(f"INFO:    filtering for start ({args.start}) and end ({args.end})")
                        fs = [f for f in fs if int(f.split('/')[-1].split('_')[1][:4]) >= int(args.start)]
                        fs = [f for f in fs if int(f.split('/')[-1].split('_')[1][-4:]) <= int(args.end)]
                        if args.verbose: print(f"INFO:    {len(fs)} files leftover")
                    files.extend(fs)
            return files

        for format in args.docformats:
            if format == 'alto':
                if args.verbose: print("INFO: Looking for ALTO files")
                paths = list({k:v for k,v in vars(args.config).items() if "ALTO" in k and v is not None}.values())
                args.alto_files = _get_files(paths, args)
                if args.verbose: print(f"INFO:    found {len(args.alto_files)} alto files")
            elif format == 'pdf':
                if args.verbose: print("INFO: Looking for PDF files")
                paths = list({k:v for k,v in vars(args.config).items() if "PDF" in k and v is not None}.values())
                args.pdf_files = _get_files(paths, args, ext=".pdf")
                if args.verbose: print(f"INFO:    found {len(args.pdf_files)} pdf files")
            elif format == 'tei':
                if args.verbose: print("INFO: Looking for TEI files")
                paths = list({k:v for k,v in vars(args.config).items() if "TEI" in k and v is not None}.values())
                args.tei_files = _get_files(paths, args)
                if args.verbose: print(f"INFO:    found {len(args.tei_files)} tei files")
        return args

    def _filter_doctype(args):
        """
        select documents based on doctype

        Args

            args: Argparse argument class instances

        Returns

            args
        """
        if args.doctypes is None:
            return args
        if args.verbose: print(f"INFO: filtering docs for type {args.doctypes}")
        for format in args.docformats:
            if hasattr(args, f"{format}_files"):
                files = getattr(args, f"{format}_files")
                if args.verbose: print(f"INFO:     starting with {len(files)} {format} files")
                files = [f for f in files if f.split('/')[-1].split('_')[0] in args.doctypes]
                if args.verbose: print(f"INFO:    --->  {len(files)} leftover")
                setattr(args, f"{format}_files", files)
        return args

    def _filter_language(args):
        """
        select documents based on the primary language

        Args

            args: Argparse argument class instances

        Returns

            args
        """
        warnings.warn(_filter_language.__name__, NotImplemented)
        return args

    def _filter_chamber(args):
        """
        select documents based on the chamber

        Args

            args: Argparse argument class instances

        Returns

            args
        """
        if args.chambers is None:
            return args
        if args.verbose: print(f"INFO: filtering docs for chamber {args.chambers}")
        for format in args.docformats:
            if hasattr(args, f"{format}_files"):
                files = getattr(args, f"{format}_files")
                if args.verbose: print(f"INFO:     starting with {len(files)} {format} files")
                files = [f for f in files if f.split('/')[-1].split('_')[2] in args.chambers]
                if args.verbose: print(f"INFO:    --->  {len(files)} leftover")
                setattr(args, f"{format}_files", files)
        return args

    args = _handle_config(args)
    args = _fetch_documents(args)
    args = _filter_doctype(args)
    args = _filter_language(args)
    args = _filter_chamber(args)

    return args
